from .calculate_centrality_measures import calculate_centrality_measures
from .calculate_cluster_distances import calculate_cluster_distances
from .calculate_cluster_sizes import calculate_cluster_sizes
from .calculate_degree_measures import calculate_degree_measures
from .calculate_distance_measures import calculate_distance_measures
from .convert_to_network import convert_to_network
from .get_depth_map import get_depth_map
from .get_neighbors_map import get_neighbors_map
from .make_voxels_array import make_voxels_array



import networkx as nx
import numpy as np
import pandas as pd
from prefect import task


@task
def calculate_centrality_measures(networks: dict) -> pd.DataFrame:
    all_measures = []

    for (seed, tick), network in networks.items():
        degree_centralities = list(nx.degree_centrality(network).values())
        closeness_centralities = list(nx.closeness_centrality(network).values())
        betweenness_centralities = list(nx.betweenness_centrality(network).values())

        degree_centrality_mean = np.mean(degree_centralities)
        closeness_centrality_mean = np.mean(closeness_centralities)
        betweenness_centrality_mean = np.mean(betweenness_centralities)

        degree_centrality_std = np.std(degree_centralities, ddof=1)
        closeness_centrality_std = np.std(closeness_centralities, ddof=1)
        betweenness_centrality_std = np.std(betweenness_centralities, ddof=1)

        all_measures.append(
            {
                "SEED": seed,
                "TICK": tick,
                "DEGREE_CENTRALITY_MEAN": degree_centrality_mean,
                "CLOSENESS_CENTRALITY_MEAN": closeness_centrality_mean,
                "BETWEENNESS_CENTRALITY_MEAN": betweenness_centrality_mean,
                "DEGREE_CENTRALITY_STD": degree_centrality_std,
                "CLOSENESS_CENTRALITY_STD": closeness_centrality_std,
                "BETWEENNESS_CENTRALITY_STD": betweenness_centrality_std,
            }
        )

    all_measures_df = pd.DataFrame(all_measures)

    return all_measures_df



from typing import Optional

import numpy as np
from prefect import task
from scipy import ndimage
from skimage import measure


@task
def get_neighbors_map(array: np.ndarray) -> dict:
    neighbors_map: dict = {cell_id: {} for cell_id in np.unique(array)}
    neighbors_map.pop(0, None)

    # Create binary mask for array.
    mask = np.zeros(array.shape, dtype="int")
    mask[array != 0] = 1

    # Label connected groups.
    labels, groups = measure.label(mask, connectivity=2, return_num=True)

    # In line function that returns a filter lambda for a given id
    voxel_filter = lambda voxel_id: lambda v: voxel_id in v

    for group in range(1, groups + 1):
        group_crop = get_cropped_array(array, group, labels)
        voxel_ids = [i for i in np.unique(group_crop) if i != 0]

        # Find neighbors for each voxel id.
        for voxel_id in voxel_ids:
            voxel_crop = get_cropped_array(group_crop, voxel_id, crop_original=True)

            # Apply custom filter to get border locations.
            border_mask = ndimage.generic_filter(voxel_crop, voxel_filter(voxel_id), size=3)

            # Find neighbors overlapping border.
            neighbor_list = np.unique(voxel_crop[border_mask == 1])
            neighbor_list = [i for i in neighbor_list if i not in [0, voxel_id]]
            neighbors_map[voxel_id] = {"group": group, "neighbors": neighbor_list}

    return neighbors_map


def get_bounding_box(array: np.ndarray) -> tuple[int, int, int, int, int, int]:
    """Finds bounding box around binary array."""
    x, y, z = array.shape

    xbounds = np.any(array, axis=(1, 2))
    ybounds = np.any(array, axis=(0, 2))
    zbounds = np.any(array, axis=(0, 1))

    xmin, xmax = np.where(xbounds)[0][[0, -1]]
    ymin, ymax = np.where(ybounds)[0][[0, -1]]
    zmin, zmax = np.where(zbounds)[0][[0, -1]]

    xmin = max(xmin - 1, 0)
    xmax = min(xmax + 2, x)

    ymin = max(ymin - 1, 0)
    ymax = min(ymax + 2, y)

    zmin = max(zmin - 1, 0)
    zmax = min(zmax + 2, z)

    return xmin, xmax, ymin, ymax, zmin, zmax


def get_cropped_array(
    array: np.ndarray, label: int, labels: Optional[np.ndarray] = None, crop_original: bool = False
) -> np.ndarray:
    # Set all voxels not matching label to zero.
    array_mask = array.copy()
    array_filter = labels if labels is not None else array_mask
    array_mask[array_filter != label] = 0

    # Crop array to label.
    xmin, xmax, ymin, ymax, zmin, zmax = get_bounding_box(array_mask)

    if crop_original:
        return array[xmin:xmax, ymin:ymax, zmin:zmax]

    return array_mask[xmin:xmax, ymin:ymax, zmin:zmax]



import pandas as pd
from prefect import task


@task
def calculate_cluster_sizes(neighbors: pd.DataFrame) -> pd.DataFrame:
    all_clusters = []

    for (seed, tick), group in neighbors.groupby(["SEED", "TICK"]):
        group_sizes = group.groupby("GROUP").size()
        cluster_sizes = group_sizes[group_sizes != 1]
        single_sizes = group_sizes[group_sizes == 1]

        all_clusters.append(
            {
                "SEED": seed,
                "TICK": tick,
                "NUM_CLUSTERS": len(cluster_sizes),
                "NUM_SINGLES": len(single_sizes),
                "CLUSTER_SIZE_TOTAL": cluster_sizes.sum(),
                "CLUSTER_SIZE_MEAN": cluster_sizes.mean(),
                "CLUSTER_SIZE_STD": cluster_sizes.std(ddof=1),
            }
        )

    all_clusters_df = pd.DataFrame(all_clusters)

    return all_clusters_df



import numpy as np
import pandas as pd
from prefect import task


@task
def calculate_degree_measures(networks: dict) -> pd.DataFrame:
    all_measures = []

    for (seed, tick), network in networks.items():
        degrees = sorted([d for n, d in network.degree()], reverse=True)
        degree_mean = np.mean(degrees)
        degree_std = np.std(degrees, ddof=1)

        all_measures.append(
            {
                "SEED": seed,
                "TICK": tick,
                "DEGREES": degrees,
                "DEGREE_MEAN": degree_mean,
                "DEGREE_STD": degree_std,
            }
        )

    all_measures_df = pd.DataFrame(all_measures)

    return all_measures_df



import networkx as nx
import numpy as np
import pandas as pd
from prefect import task


@task
def calculate_distance_measures(networks: dict) -> pd.DataFrame:
    all_measures = []

    for (seed, tick), network in networks.items():
        if not nx.is_connected(network):
            subnetworks = [
                network.subgraph(component) for component in nx.connected_components(network)
            ]

            radius = np.mean([nx.radius(subnetwork) for subnetwork in subnetworks])
            diameter = np.mean([nx.diameter(subnetwork) for subnetwork in subnetworks])

            eccentricities = [nx.eccentricity(subnetwork) for subnetwork in subnetworks]
            eccentricity = np.mean([np.mean(list(ecc.values())) for ecc in eccentricities])

            shortest_paths = [
                nx.average_shortest_path_length(subnetwork) for subnetwork in subnetworks
            ]
            shortest_path = np.mean(shortest_paths)
        else:
            radius = nx.radius(network)
            diameter = nx.diameter(network)

            ecc = nx.eccentricity(network)
            eccentricity = np.mean(list(ecc.values()))

            shortest_path = nx.average_shortest_path_length(network)

        all_measures.append(
            {
                "SEED": seed,
                "TICK": tick,
                "RADIUS": radius,
                "DIAMETER": diameter,
                "ECCENTRICITY": eccentricity,
                "SHORTEST_PATH": shortest_path,
            }
        )

    all_measures_df = pd.DataFrame(all_measures)

    return all_measures_df



import numpy as np
import pandas as pd
from prefect import task
from scipy.spatial import distance


@task
def calculate_cluster_distances(neighbors: pd.DataFrame) -> pd.DataFrame:
    all_clusters = []

    for (seed, tick), group in neighbors.groupby(["SEED", "TICK"]):
        clusters = group.groupby("GROUP").filter(lambda x: len(x.index) > 1)

        if clusters.empty:
            continue

        inter_distance_mean, inter_distance_std = calculate_inter_cluster_distances(clusters)
        intra_distance_mean, intra_distance_std = calculate_intra_cluster_distances(clusters)

        all_clusters.append(
            {
                "SEED": seed,
                "TICK": tick,
                "INTER_DISTANCE_MEAN": inter_distance_mean,
                "INTER_DISTANCE_STD": inter_distance_std,
                "INTRA_DISTANCE_MEAN": intra_distance_mean,
                "INTRA_DISTANCE_STD": intra_distance_std,
            }
        )

    all_clusters_df = pd.DataFrame(all_clusters)

    return all_clusters_df


def calculate_inter_cluster_distances(clusters: pd.DataFrame) -> tuple[float, float]:
    cluster_centroids = clusters.groupby("GROUP")[["cx", "cy", "cz"]].mean().values

    if cluster_centroids.shape[0] < 2:
        return (np.nan, np.nan)

    inter_distances = distance.cdist(cluster_centroids, cluster_centroids, "euclidean")
    distances = np.ndarray.flatten(inter_distances)
    distances = np.delete(distances, range(0, len(distances), len(inter_distances) + 1), 0)

    inter_distance_mean = np.mean(distances)
    inter_distance_std = np.std(distances, ddof=1)

    return (inter_distance_mean, inter_distance_std)


def calculate_intra_cluster_distances(clusters: pd.DataFrame) -> tuple[float, float]:
    intra_distance_means = []
    intra_distance_stds = []

    for _, group in clusters.groupby("GROUP"):
        cluster_centroids = group[["cx", "cy", "cz"]].values

        intra_distances = distance.cdist(cluster_centroids, cluster_centroids, "euclidean")
        distances = np.ndarray.flatten(intra_distances)
        distances = np.delete(distances, range(0, len(distances), len(intra_distances) + 1), 0)

        intra_distance_means.append(np.mean(distances))
        intra_distance_stds.append(np.std(distances, ddof=1))

    return (np.mean(intra_distance_means), np.mean(intra_distance_stds))



import numpy as np
from prefect import task


@task
def make_voxels_array(locations: dict) -> np.ndarray:
    # Extract all voxel positions with id.
    all_ids: list[int] = []
    all_xyz: list[tuple[int, int, int]] = []
    for location in locations:
        cell_id = location["id"]
        xyz = [(x, y, z) for region in location["location"] for x, y, z in region["voxels"]]
        all_xyz = all_xyz + xyz
        all_ids = all_ids + [cell_id] * len(xyz)

    # Create empty array.
    mins = np.min(all_xyz, axis=0)
    maxs = np.max(all_xyz, axis=0)
    length, width, height = np.subtract(maxs, mins) + 3
    array = np.zeros((height, width, length), dtype=np.uint16)

    # Return if no voxels.
    if len(all_ids) == 0:
        return array

    # Fill voxel array.
    all_xyz_offset = [(z - mins[2] + 1, y - mins[1] + 1, x - mins[0] + 1) for x, y, z in all_xyz]
    array[tuple(np.transpose(all_xyz_offset))] = all_ids

    return array



import numpy as np
from prefect import task
from scipy import ndimage
from skimage import measure


@task
def get_depth_map(array: np.ndarray, neighbors_map: dict) -> dict:
    depth_map = {cell_id: 0 for cell_id in np.unique(array)}
    depth_map.pop(0, None)

    edge_ids = find_edge_ids(array)
    visited = set(edge_ids)
    queue = edge_ids.copy()

    while queue:
        current_id = queue.pop(0)

        current_neighbors = neighbors_map[current_id]["neighbors"]
        valid_neighbors = set(current_neighbors) - visited
        visited.update(valid_neighbors)
        queue = queue + list(valid_neighbors)

        for neighbor_id in valid_neighbors:
            depth_map[neighbor_id] = depth_map[current_id] + 1

        depth_map[current_id] = depth_map[current_id] + 1

    return depth_map


def find_edge_ids(array: np.ndarray) -> list[int]:
    slice_index = np.argmax(np.count_nonzero(array, axis=(1, 2)))
    array_slice = array[slice_index, :, :]

    # Calculate voronoi from cell shapes.
    distances = ndimage.distance_transform_edt(
        array_slice == 0, return_distances=False, return_indices=True
    )
    distances = distances.astype("uint16", copy=False)
    coordinates_y = distances[0].flatten()
    coordinates_x = distances[1].flatten()
    voronoi = array_slice[coordinates_y, coordinates_x].reshape(array_slice.shape)

    # Create border mask.
    mask = np.zeros(array_slice.shape, dtype="uint8")
    mask[array_slice != 0] = 1
    while measure.euler_number(mask) != 1:
        mask = ndimage.binary_dilation(mask, iterations=1)

    # Filter voronoi by mask to get edge ids.
    voronoi[mask == 1] = 0
    edge_ids = list(np.unique(voronoi))
    edge_ids.remove(0)

    return edge_ids



import networkx as nx
import pandas as pd
from prefect import task


@task
def convert_to_network(neighbors: pd.DataFrame) -> nx.Graph:
    networks = {}

    for tick, group in neighbors.groupby("TICK"):
        nodes = list(group["ID"].values)
        edges = [
            (node_id, neighbor_id)
            for node_id, neighbor_ids in zip(group["ID"], group["NEIGHBORS"])
            for neighbor_id in neighbor_ids
        ]

        network = nx.Graph()
        network.add_nodes_from(nodes)
        network.add_edges_from(edges)

        networks[tick] = network

    return networks
